/* 
   qtfile_core.c (C) Justin Schoeman 2001 (justin@suntiger.ee.up.ac.za)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
*/

#include <string.h>
#include <quicktime/quicktime.h>
#include "RTjpeg/RTjpeg.h"
#include "NVrec.h"

struct qtfile_core_t {
	quicktime_t * fd;
	RTjpeg_t * rtj;
	unsigned char * cbuf;
	int abits;
	int achans;
	char compr[5];
	int width;
	int height;
};

struct file_core_t qtfile_core;


void qtfile_core_close(struct file_core_t * fct)
{
	struct qtfile_core_t * qtc = (struct qtfile_core_t *)fct->priv;
	if(!qtc)
		return;
	RTjpeg_close(qtc->rtj);
	quicktime_close(qtc->fd);
	free(qtc->cbuf);
	free(qtc);
	free(fct);
}

void qtfile_core_nullvideo(struct file_core_t * fct)
{
	struct qtfile_core_t * qtc = (struct qtfile_core_t *)fct->priv;
	int i;
	
	if(!qtc)
		return;
//	fprintf(stderr, "file core null video\n");
	i = RTjpeg_nullcompress(qtc->rtj, qtc->cbuf);
	quicktime_write_frame(qtc->fd, qtc->cbuf, i, 0);
}

void qtfile_core_video(struct file_core_t * fct, unsigned char **buf)
{
	struct qtfile_core_t * qtc = (struct qtfile_core_t *)fct->priv;
	int i;
	
	if(!qtc)
		return;
//	fprintf(stderr, "file core video\n");
	i = RTjpeg_compress(qtc->rtj, qtc->cbuf, buf); //planes);
	quicktime_write_frame(qtc->fd, qtc->cbuf, i, 0);
}

/* FIXME - not sure if this will work ... */

void qtfile_core_nullvideo_raw(struct file_core_t * fct)
{
	struct qtfile_core_t * qtc = (struct qtfile_core_t *)fct->priv;
//	int i;
	
	if(!qtc)
		return;
	quicktime_write_frame(qtc->fd, qtc->cbuf, 0, 0);
}

void qtfile_core_video_raw(struct file_core_t * fct, unsigned char **buf)
{
	struct qtfile_core_t * qtc = (struct qtfile_core_t *)fct->priv;
	int i, j, k = 0;
	
	if(!qtc)
		return;
	for(i = 0; i < qtc->height; i++) {
		for(j = 0; j < qtc->width; j++) {
			qtc->cbuf[k++] = buf[i][j * 3 + 2];
			qtc->cbuf[k++] = buf[i][j * 3 + 1];
			qtc->cbuf[k++] = buf[i][j * 3 + 0];
		}
	}
	quicktime_write_frame(qtc->fd, qtc->cbuf, qtc->width * qtc->height * 3, 0);
}

void qtfile_core_video_yuv2(struct file_core_t * fct, unsigned char **buf)
{
	struct qtfile_core_t * qtc = (struct qtfile_core_t *)fct->priv;
	int i, j;
	unsigned int * src, * dst;
	
	if(!qtc)
		return;
	dst = (unsigned int *)qtc->cbuf; 
	for(i = 0; i < qtc->height; i++) {
		src = (unsigned int *)(buf[i]);
		for(j = 0; j < (qtc->width / 2); j++) {
			*(dst++) = src[j] ^ 0x80008000;
		}
	}
	quicktime_write_frame(qtc->fd, qtc->cbuf, qtc->width * qtc->height * 2, 0);
}

void qtfile_core_audio(struct file_core_t * fct, unsigned char *buf, int n)
{
	struct qtfile_core_t * qtc = (struct qtfile_core_t *)fct->priv;
	unsigned short *bs = (unsigned short *)buf;
	int i, s;

	if(!qtc)
		return;
	
	s = n >> ((qtc->abits == 16) ? 1 : 0);
	if(qtc->abits == 16) {
		for(i = 0; i < s; i++)
			bs[i] = ((bs[i]>>8)&0x00ff)|((bs[i]<<8)&0xff00);
	} else {
		for(i = 0; i < s; i++)
			buf[i] ^= 0x80;
	}

	s >>= (qtc->achans == 2) ? 1 : 0;
	quicktime_write_audio(qtc->fd, buf, s, 0);
//	fprintf(stderr, "file core audio %d %d %d\n", n, s, buf[1]);
}

void qtfile_core_sync(struct file_core_t * fct)
{
	struct qtfile_core_t * qtc = (struct qtfile_core_t *)fct->priv;
	fdatasync(fileno(qtc->fd->stream));
}

int qtfile_core_format(char **args)
{
	int fmt = NV_YUV420;
 	NV_ARGS;
 	
	NV_SCAN_ARGS {
		if(strcmp(NV_THIS_ARG, "-raw") == 0) {
			fmt = NV_RGB24;
		} else if(strcmp(NV_THIS_ARG, "-yuv2") == 0) {
			fmt = NV_YUV422;
		} 
	}
	return fmt;
}

void qtfile_core_usage()
{
	fprintf(stderr, "quicktime file core usage:\n");
	fprintf(stderr, "\t-k <key frame rate>  [25]\n");
	fprintf(stderr, "\t-q <quality factor>  [255]\n");
	fprintf(stderr, "\t-l <luma factor>     [2]\n");
	fprintf(stderr, "\t-c <chroma factor>   [1]\n");
	fprintf(stderr, "\t-o <output filename> [NONE]\n");
}

struct file_core_t * qtfile_core_init(char ** args, ntime_t vrate)
{
	struct file_core_t * fct;
	struct qtfile_core_t * qtc;
	double f;
	int i;
 	int arate = -1;
 	char *fname = NULL;
 	int keyrate = 25;
 	int qfactor = 255;
 	int lfactor = 2;
 	int cfactor = -1;
 	NV_ARGS;
 	
 	fct =  malloc(sizeof(struct file_core_t));
 	if(!fct) {
 		perror("qtfile core init - could no malloc new cap core");
 		return NULL;
 	}
 	qtc = malloc(sizeof(struct qtfile_core_t));
 	if(!qtc) {
 		perror("qtfile core init - could no malloc new qt core");
 		free(fct);
 		return NULL;
 	}
 	memcpy(fct, &qtfile_core, sizeof(struct file_core_t));
 	fct->priv = (void *)qtc;
 	
	NV_SCAN_ARGS {
		if(strcmp(NV_THIS_ARG, "-r") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			arate = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-b") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			qtc->abits = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-s") == 0) {
			qtc->achans = 2;
		} else if(strcmp(NV_THIS_ARG, "-m") == 0) {
			qtc->achans = 1;
		} else if(strcmp(NV_THIS_ARG, "-w") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			qtc->width = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-h") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			qtc->height = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-k") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			keyrate = atoi(NV_THIS_ARG);
			if(keyrate < 0)
				goto usage;
		} else if(strcmp(NV_THIS_ARG, "-q") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			qfactor = atoi(NV_THIS_ARG);
			if((qfactor < 16) || (qfactor > 255))
				goto usage;
		} else if(strcmp(NV_THIS_ARG, "-l") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			lfactor = atoi(NV_THIS_ARG);
			if(lfactor < 0)
				goto usage;
		} else if(strcmp(NV_THIS_ARG, "-c") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			cfactor = atoi(NV_THIS_ARG);
			if(cfactor < 0)
				goto usage;
		} else if(strcmp(NV_THIS_ARG, "-o") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			fname = NV_THIS_ARG;
		} else if(strcmp(NV_THIS_ARG, "-raw") == 0) {
			strcpy(qtc->compr, QUICKTIME_RAW);
			fct->file_core_nullvideo = qtfile_core_nullvideo_raw;
			fct->file_core_fastvideo = qtfile_core_video_raw;
			fct->file_core_video = qtfile_core_video_raw;
		} else if(strcmp(NV_THIS_ARG, "-yuv2") == 0) {
			strcpy(qtc->compr, QUICKTIME_YUV2);
			fct->file_core_nullvideo = qtfile_core_nullvideo_raw;
			fct->file_core_fastvideo = qtfile_core_video_yuv2;
			fct->file_core_video = qtfile_core_video_yuv2;
		} 
	}
	if(!qtc->compr[0]) {
		strcpy(qtc->compr, "RTJ0");
		if((qtc->width&15) || (qtc->height&15))
			goto usage;
	}
	if(cfactor < 0)
		cfactor = lfactor / 2;
	if(!fname)
		goto usage;
	fprintf(stderr, "file core init - file: %s\n", fname);
	
	unlink(fname);
	qtc->fd = quicktime_open(fname, 1, 1);
	if(!qtc->fd) {
		perror("file core init - could not open file");
		goto error0;
	}
	f = 1e9 / (double)vrate;
	fprintf(stderr, "file core init - frame rate is %f\n", f);
	quicktime_set_video(qtc->fd, 1, qtc->width, qtc->height, f, qtc->compr);
	quicktime_set_audio(qtc->fd, qtc->achans, arate, qtc->abits, QUICKTIME_TWOS);

	qtc->rtj = RTjpeg_init();
	if(!qtc->rtj)
		goto error1;
	RTjpeg_set_size(qtc->rtj, &qtc->width, &qtc->height);
	RTjpeg_set_quality(qtc->rtj, &qfactor);
	i = RTJ_YUV420;
	RTjpeg_set_format(qtc->rtj, &i);
	RTjpeg_set_intra(qtc->rtj, &keyrate, &lfactor, &cfactor);
	
	qtc->cbuf = malloc(qtc->width * qtc->height * 4);
	if(!qtc->cbuf) {
		perror("file core init - could not allocate compression buffer");
		goto error2;
	}
	return (void *)fct;

error2:
	RTjpeg_close(qtc->rtj);
error1:
	quicktime_close(qtc->fd);
error0:
	free(fct);
	free(qtc);
	return NULL;
usage:
	fprintf(stderr, "Error parsing options.  Please run without any command line options for\n");
	fprintf(stderr, "detailed usage information.\n");
	free(fct);
	free(qtc);
	return NULL;
}

struct file_core_t qtfile_core = {
	file_core_init: qtfile_core_init,
	file_core_close: qtfile_core_close,
	file_core_nullvideo: qtfile_core_nullvideo,
	file_core_fastvideo: qtfile_core_video,
	file_core_video: qtfile_core_video,
	file_core_audio: qtfile_core_audio,
	file_core_sync: qtfile_core_sync,
	file_core_format: qtfile_core_format,
	file_core_usage: qtfile_core_usage,
	priv: NULL
};
